use std::env::args;
use std::ffi::OsString;
use std::fs::File;
use std::io::{Read, Seek, Write};
use std::iter::zip;
use std::path::Path;
use std::process::Command;
use std::str::FromStr;

use tempfile::NamedTempFile;

use enemize::Patch;

fn main() -> Result<(), anyhow::Error> {
    println!("BinComp ---- ");

    let mut args = args();

    let symbols_fn = args.next_back().expect("No symbols filename");
    let base_patch = args.next_back().expect("No basePatch filename");
    let asm_fn = args.next_back().expect("No input ASM filename");

    let mut zeroes_patch = vec![];

    let mut zeroes_file = NamedTempFile::new()?;
    let mut ones_file = NamedTempFile::new()?;

    println!(
        "temp file (zeros): {}",
        zeroes_file.path().to_string_lossy()
    );
    println!("temp file (0xFF): {}", ones_file.path().to_string_lossy());

    run_asar(asm_fn.as_ref(), zeroes_file.path(), symbols_fn.as_ref());

    println!("Reading all 0x00 file.");
    zeroes_file.read_to_end(&mut zeroes_patch)?;
    println!("Read all 0x00 file.");

    println!("Generating all 0xFF file.");
    let all_ones = vec![0xff; zeroes_patch.len()];
    ones_file.write_all(&all_ones)?;
    ones_file.flush()?;
    println!("Generated all 0xFF file.");

    run_asar(asm_fn.as_ref(), ones_file.path(), symbols_fn.as_ref());

    println!("Reading all 0xFF patched file.");
    let mut ones_patch = vec![];
    ones_file.rewind()?;
    ones_file.read_to_end(&mut ones_patch)?;
    println!("Read all 0xFF patched file.");

    assert!(
        zeroes_patch.len() == ones_patch.len(),
        "File lengths don't match (0: {}, 1: {})! Aborting.",
        zeroes_patch.len(),
        ones_patch.len()
    );

    println!("");

    let (_, list) = zip(zeroes_patch, ones_patch).enumerate().fold((true, Vec::new()), |(start, mut list), (i, (zero, ff))| {
        if zero == ff {
            if start {
                println!("New patch data found at address {:X}", i);
                let patch = Patch { address: i, patch_data: vec![] };
                list.push(patch);
            }
            list.last_mut().and_then(|p| {
                p.patch_data.push(zero);
                Some(p)
            });

            (false, list)
        } else {
            assert!(zero == 0x00 && ff == 0xff, "Something went wrong. Zero file has non-zero or FF file has non-FF where files do not match!");

            (true, list)
        }
    });

    println!("Writing patch data file.");
    let bp_file = File::create(&base_patch)?;
    serde_json::to_writer(bp_file, &list)?;

    println!("Cleaning up temp files.");
    println!("BinComp finished ----");

    Ok(())
}

fn run_asar(asm: &Path, rom: &Path, symbols: &Path) {
    println!("---- Running asar --no-title-check --fix-checksum=off --symbols=wla --symbols-path={} {} {} ----", symbols.to_string_lossy(), asm.to_string_lossy(), rom.to_string_lossy());

    let mut symbols_path_arg = OsString::from_str("--symbols-path=").unwrap();
    symbols_path_arg.push(symbols);

    let status = Command::new("asar")
        .arg("--no-title-check")
        .arg("--fix-checksum=off")
        .arg("--symbols=wla")
        .arg(symbols_path_arg)
        .arg(asm)
        .arg(rom)
        .status()
        .expect("Failed to run asar");

    assert!(status.success(), "asar threw errors");

    println!("---- asar finished ----");
}
